
#include "si5351/si5351.h"
#include "ds3231/ds3231.h"
#include "utils/symbol_rate.h"
#include "utils/command_buffer.h"
#include "jt_encoder/jt_wspr_encoder.h"
#include "jt_encoder/jt_mock_encoder.h"
#include "jt_encoder/jt_jt65_encoder.h"
#include "jt_encoder/jt_iscat_encoder.h"
#include "utils/jt_band_params.h"

#include <Wire.h>
#include <avr/pgmspace.h>

// ---------------- Global Variables -----------------

Si5351 si5351(0);
SymbolRate symbolRate;
CCommandBuffer commandBuffer;

JTEncodeBase::iterator itSymbol;
JTEncodeBase::iterator itEnd;

// To save dynamic and program space you may comment unnecessary coders and comment references to them in the code.
JTWSPREncoder wsprCoder;

unsigned prevSymbolIndex;
unsigned char prevSymbolValue;
JTBandParams bandParams;
enum State
{
  stateIdle = 0,
  stateWaitStart,
  statePTTActive,
  stateTransmitting
};
int currentState = stateIdle;


RtcDatetime pttStartTime;
RtcDatetime symbolsStartTime;

int prev1PPS;
int numShowTime = 0;


//-=-=-=-=-=-=-=-=-=-=-=-=-=-=--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-

// ---- Arduino pin configuration ---

const int pinLED = LED_BUILTIN; // duplicate 1pps
const int pin1PPS = 4; // <-- 1pps signal from ds3231


// ----------- Startup Parameters  ------------------------------

const int PTTWarmupTimeInSeconds = 2; // Number of seconds prior symbols transmission. It is for activate PTT of Power Amplifer.
const unsigned TRIntervalMultiplier = 1; // for example: value 1 means TR-TR-TR...   value 3 means: TX-RX-RX-TX...
bool WorkModeActive = true; // active TX logic at power-on


const size_t currentBandIndex = 0;

//
// Copy vector-descriptor from web configurator and paste into bandDescrArray.
// Local configurator:  ./doc/band_configurator.html 
// Configurator in the Net:  http://ra9yer.blogspot.com/p/si5351-configurator.html
//

JTBandDescr bandDescrArray[] = {
  {Mode_WSPR2, 31, 154287, 614418, 30, 1, 12, 8192, 120} // f=28.126 MHz; WSPR2; step=1.465Hz; 1.465baud; T/R=2m};
};

//----------  End Configuration ------------------------



// ------ Message initialize functions ----- 

//--------------------------------------------------
void initializeWSPRCoder()
{
  wsprCoder.encodeMessage( "AA0AAA", "NO13", 10 );    // <---- WSPR message defined here.
}


//==============================


//--------------------------------------------------
void setup() {
  // put your setup code here, to run once:
  Serial.begin(9600);
  Wire.begin();

  pinMode(pinLED, OUTPUT);
  pinMode(pin1PPS, INPUT_PULLUP);

  Ds3231::enable1PPS(true);
  
  si5351.initialize();
  si5351.enableOutput(Si5351::OUT_0, false);

  initializeWSPRCoder();

  bandParams.initFromJTBandDescr( bandDescrArray[currentBandIndex] ); 
  printBandInfo();

  adjustLaunchTimeIfNeed();

}


//--------------------------------------------------
void loop() {

  //
  // command line interface
  //
  if( Serial.available() )
  {
    int inByte = Serial.read();
    commandBuffer.handleInputChar( inByte );

    if( commandBuffer.endsWith("\r") || commandBuffer.endsWith("\n") )
    {
      processTerminalCommands();
      commandBuffer.reset();
    }
  }

  //
  // sending symbols
  //
  if( currentState == stateTransmitting )
  {
    unsigned currentSymbolIndex = symbolRate.currentSymbolIndex();
  
    if( currentSymbolIndex != prevSymbolIndex )
    {
      unsigned tmpIndex = prevSymbolIndex;
      // Step to next symbol.
      // In some very rarely case we may skip a few symbols bacause symboldIndex is more relevant parameter.
      while( tmpIndex < currentSymbolIndex )
      {
        if( itSymbol != itEnd )
        {
          ++itSymbol;
        }
        ++tmpIndex;
      }

      if( itSymbol != itEnd )
      {
        unsigned char currentSymbolValue = *itSymbol;
        if( prevSymbolValue != currentSymbolValue )
        {
          setSymbol( currentSymbolValue );
        }
        prevSymbolValue = currentSymbolValue;

        if( currentSymbolIndex % 5 == 0 )
        {
          // just to view progress
          Serial.println(currentSymbolIndex);
        }
      }
      else
      {
        Serial.println(F("Done"));
        stop_tx();
        adjustLaunchTimeIfNeed();
      }

    }
    prevSymbolIndex = currentSymbolIndex;
  }

  //
  // detect next second
  //
  int current1PPS = digitalRead(pin1PPS);
  digitalWrite(pinLED, current1PPS); // duplicate LED
  bool nextSecondDetected = ( prev1PPS && !current1PPS );
  prev1PPS = current1PPS;

  // show time when cammand: ts N
  if( nextSecondDetected && numShowTime>0 )
  {
    printTime(false);
    --numShowTime;
  }

  //
  // actions when next second event is detected
  //
  if( nextSecondDetected &&  (currentState == stateWaitStart || currentState == statePTTActive) )
  {
    RtcDatetime currentTime;
    Ds3231::getTime(currentTime);

    if( currentState == stateWaitStart && currentTime >= pttStartTime )
    {
      Serial.print(F("PTT active\n"));
      printTime(false);
      currentState = statePTTActive;
      activate_ptt();
    }
    
    if( currentState == statePTTActive && currentTime >= symbolsStartTime )
    {
      Serial.print(F("Sending data active\n"));
      printTime(false);
      currentState = stateTransmitting;
      start_tx();
    }
  }

  

}

//--------------------------------------------------
void initCoderFromProgmemArray(JTMockEncoder& coder, const unsigned char* srcProgMemPtr, unsigned len, unsigned pitsPerSymbol)
{
  const int numberOfSymbolsLargeEnough = 206;  // jt4 has 206symbols - more than others.
  unsigned char buf[numberOfSymbolsLargeEnough];
  memcpy_P(buf, srcProgMemPtr, len);  // copy from progmem to local array
  coder.initFromArray(buf, len, pitsPerSymbol);
}



//--------------------------------------------------
void processTerminalCommands()
{
  if( commandBuffer.startsWith(F("setdate")) )  // update datetime in ds3231 chip
  {
    // process set date
    unsigned params[6];
    memset(params, 0, sizeof(params));
    for(size_t i=0; i<6; i++ )
    {
      const char * ptr = commandBuffer.getNextToken(" :-");
      if(ptr == NULL) {
        break;
      }
      params[i] = atoi(ptr);
    }
    
    RtcDatetime time;
    time.second = params[5];
    time.minute = params[4];
    time.hour = params[3];
    time.day = params[2];
    time.month = params[1];
    time.year = params[0] % 100;
    Ds3231::setTime(time);

    Serial.print(F("\nDate updated\n"));

  }
  else if( commandBuffer.startsWith(F("ts")) ) // time show [N times]
  {
    // show time command
    const char * ptr = commandBuffer.getNextToken(" "); // read first argument
    if(ptr == NULL) {
      // simple show time
      printTime(false);
    }
    else
    {
      // show time N times with 1sec interval
      numShowTime = atoi(ptr);
    }
    
  }
  else if( commandBuffer.startsWith(F("start")) ) // manual start
  {
    if( currentState != stateTransmitting )
    {
      activate_ptt();
      start_tx();
    }
    else
    {
      Serial.println(F("unable to start when tx active"));
    }
    
  }
  else if( commandBuffer.startsWith(F("stop")) )  // manual stop
  {
    stop_tx();
    if( WorkModeActive )
    {
      adjustLaunchTimeIfNeed();
    }
    else
    {
      currentState = stateIdle; 
    }
  }
  else if( commandBuffer.startsWith(F("activate")) )  // activate auto start mode
  {
    Serial.print(F("Activated"));
    WorkModeActive = true;
    if( currentState != stateTransmitting )
    {
      adjustLaunchTimeIfNeed();
    }
  }
  else if( commandBuffer.startsWith(F("deactivate")) ) // deactivate auto start mode
  {
    Serial.print(F("Deactivated"));
    WorkModeActive = false;
    pttStartTime.reset();
    symbolsStartTime.reset();
    if( currentState != stateTransmitting )
    {
      currentState = stateIdle;
    }
    
  }
  else if( commandBuffer.startsWith(F("si")) )  // system info
  {
    printTime(true);
    printBandInfo();
  }
  else if( commandBuffer.startsWith(F("temperature")) )
  {
    int16_t temperature = Ds3231::getTemperature();
    Serial.print(F("temperature="));
    Serial.print(temperature);
    Serial.println();
  }
  
}

//--------------------------------------------------
void printTime(bool includeLaunchTime)
{
  char buf[20];

  RtcDatetime time;
  Ds3231::getTime(time);
  Serial.print(F("\nClock  DateTime = "));
  time.formatStr(buf);
  Serial.print(buf);

  if( includeLaunchTime )
  {
    Serial.print(F("\nPTT    DateTime = "));
    pttStartTime.formatStr(buf);
    Serial.print(buf);

    Serial.print(F("\nSymbol DateTime = "));
    symbolsStartTime.formatStr(buf);
    Serial.print(buf);
  }

  Serial.print(F("\n"));
}

//--------------------------------------------------
void printBandInfo()
{
  Serial.print(F("\nCurrent band: Index="));
  Serial.print(currentBandIndex);
  Serial.print(F("; mode="));
  Serial.print((int)bandParams.getJTMode());
  Serial.print(F("; F="));
  Serial.print(bandParams.approxFrequencyInMHz());
  Serial.print(F(" MHz"));
  Serial.print(F("\n"));
}

//----------------------------------------------------------
void setSymbol(unsigned symbol)
{
  uint16_t a;
  uint32_t b;
  uint32_t c;
  bandParams.getPLLParamsForSymbol(symbol, a, b, c);
  si5351.setupPLLParams(Si5351::PLL_A, a, b, c);
}


//----------------------------------------------------------
void start_tx()
{
  Serial.println(F("Started"));
  printBandInfo();

  switch( bandParams.getJTMode() )
  {
    case Mode_WSPR2:
    case Mode_WSPR15:
      // wsprCoder was already initialized at setup()
      itSymbol = wsprCoder.begin();
      itEnd = wsprCoder.end();
      break;
      
    default:
      Serial.println(F("Unsupported encoder"));
      stop_tx();
      currentState = stateIdle;
      return;
  }

  if( itSymbol == itEnd )
  {
    Serial.println(F("No symbols to transmit"));
    stop_tx();
    currentState = stateIdle;
    return;
  }


  si5351.setupMultisyncParams(Si5351::OUT_0, bandParams.getMsyncDiv(), bandParams.getRDiv() );
  
  unsigned char currentSymbolValue = *itSymbol;         
  setSymbol( currentSymbolValue );
  
  prevSymbolValue = currentSymbolValue;
  prevSymbolIndex = 0;
  
  currentState = stateTransmitting;
  
  symbolRate.initFromParams( bandParams.getBaudRateDividend(), bandParams.getBaudRateDivisor() );
  symbolRate.resetToNow();
  
  si5351.enableOutput(Si5351::OUT_0, true); //
}

//----------------------------------------------------------
void stop_tx()
{
  Serial.println(F("Stopped"));
  si5351.enableOutput(Si5351::OUT_0, false);
  deactivate_ptt();
}


//----------------------------------------------------------
void adjustLaunchTimeIfNeed()
{
  if( !WorkModeActive )
  {
    currentState = stateIdle;
    return;
  }
  
  RtcDatetime currentTime;
  if( Ds3231::getTime(currentTime) != Ds3231::rcSuccess )
  {
    Serial.print(F("\nUnable to initialize ds3231"));
    currentState = stateIdle;
  }
  
  unsigned trInterval = bandParams.getTRInterval();
  unsigned long stamp = currentTime.stamp();
  stamp += trInterval * TRIntervalMultiplier; // jump to future
  stamp -= (stamp % trInterval); // round down
  stamp += 1; // JT65,JT9,JT4,WSPR - all start at 01 second 

  symbolsStartTime.initFromStamp( stamp );
  pttStartTime.initFromStamp( stamp - PTTWarmupTimeInSeconds );

  printTime(true);

  currentState = stateWaitStart;
  
}




//----------------------------------------------------------
void activate_ptt()
{

  // turn on ptt key here, if the feature present.
  
}

//----------------------------------------------------------
void deactivate_ptt()
{
  
  // turn off ptt key here
}


